# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017 Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Profiling results as a table"""

import logging
import os.path
from ui.qt import (Qt, pyqtSignal, QCursor,
                   QTreeWidgetItem, QTreeWidget, QLabel, QWidget, QVBoxLayout,
                   QHeaderView, QMenu, QAbstractItemView, QSizePolicy)
from ui.itemdelegates import NoOutlineHeightDelegate
from ui.labels import HeaderFitLabel
from utils.globals import GlobalData
from utils.pixmapcache import getIcon


FLOAT_FORMAT = "%8.6f"
MAX_CALLS_IN_TOOLTIP = 32

OUTSIDE_COL_INDEX = 0
CALLS_COL_INDEX = 1
TOTAL_COL_INDEX = 2
TOTALPERCALL_COL_INDEX = 3
CUM_COL_INDEX = 4
CUMPERCALL_COL_INDEX = 5
LOCATION_COL_INDEX = 6
NAME_COL_INDEX = 7
CALLERS_COL_INDEX = 8
CALLEES_COL_INDEX = 9


class ProfilingTableItem(QTreeWidgetItem):

    """Profiling table row"""

    def __init__(self, items, isOutside, funcIDs):
        QTreeWidgetItem.__init__(self, items)

        self.__isOutside = isOutside
        self.__funcIDs = funcIDs

        # Set the first column icon
        if isOutside:
            self.setIcon(OUTSIDE_COL_INDEX, getIcon('nonprojectentry.png'))
            self.setToolTip(OUTSIDE_COL_INDEX,
                            'Record of an outside function')
        else:
            self.setIcon(OUTSIDE_COL_INDEX, getIcon('empty.png'))
            self.setToolTip(OUTSIDE_COL_INDEX, '')

        # Set the function name tooltip
        fileName = self.getFileName()
        lineNumber = self.getLineNumber()
        if fileName != "" and lineNumber != 0:
            self.setToolTip(NAME_COL_INDEX,
                            GlobalData().getFileLineDocstring(fileName,
                                                              lineNumber))

        # Sets the location/name columns
        self.updateLocation(False)
        self.setText(NAME_COL_INDEX, self.getFunctionName())

        for column in [CALLS_COL_INDEX, TOTALPERCALL_COL_INDEX,
                       CUM_COL_INDEX, CUMPERCALL_COL_INDEX,
                       CALLERS_COL_INDEX, CALLEES_COL_INDEX]:
            self.setTextAlignment(column, Qt.AlignRight)
        self.setTextAlignment(TOTAL_COL_INDEX, Qt.AlignLeft)

    def isOutside(self):
        """True if it is outside"""
        return self.__isOutside

    def getFuncIDs(self):
        """Provides the function identification exactly as pstats defines it"""
        return self.__funcIDs

    def getFileName(self):
        """Provides the item file name"""
        if self.__funcIDs[0] == '~':
            return ""
        return self.__funcIDs[0]

    def getLineNumber(self):
        """Provides the line number"""
        return self.__funcIDs[1]

    def getFunctionName(self):
        """Provides the function name"""
        name = self.__funcIDs[2]
        if self.__funcIDs[:2] == ('~', 0):
            # special case for built-in functions
            if name.startswith('<') and name.endswith('>'):
                return "{%s}" % name[1:-1]
        return name

    def updateLocation(self, isFull):
        """Updates the function location cell"""
        fileName = self.getFileName()
        if fileName != "":
            if isFull:
                loc = fileName + ":" + str(self.getLineNumber())
            else:
                loc = os.path.basename(fileName) + ":" + \
                      str(self.getLineNumber())
            self.setText(LOCATION_COL_INDEX, loc)
            self.setToolTip(LOCATION_COL_INDEX,
                            fileName + ":" + str(self.getLineNumber()))

    def callersCount(self):
        """Provides the number of callers"""
        return int(str(self.text(CALLERS_COL_INDEX)))

    def calleesCount(self):
        """Provides the number of callees"""
        return int(str(self.text(CALLEES_COL_INDEX)))

    @staticmethod
    def __getActualCalls(txt):
        """Returns the actual number of calls as integer"""
        return int(str(txt).split('/', 1)[0])

    @staticmethod
    def __getFloatValue(txt):
        """Returns a float value from a column"""
        return float(str(txt).split(' ', 1)[0])

    def __lt__(self, other):
        """Integer or string sorting"""
        sortColumn = self.treeWidget().sortColumn()

        if sortColumn == OUTSIDE_COL_INDEX:
            return self.__isOutside < other.__isOutside

        txt = self.text(sortColumn)
        otherTxt = other.text(sortColumn)
        if sortColumn == CALLS_COL_INDEX:
            return self.__getActualCalls(txt) < \
                   self.__getActualCalls(otherTxt)

        if sortColumn in [TOTAL_COL_INDEX, TOTALPERCALL_COL_INDEX,
                          CUM_COL_INDEX, CUMPERCALL_COL_INDEX]:
            return self.__getFloatValue(txt) < \
                   self.__getFloatValue(otherTxt)

        if sortColumn == LOCATION_COL_INDEX:
            sfName = self.getFileName()
            ofName = other.getFileName()
            if sfName == ofName:
                return self.getLineNumber() < other.getLineNumber()
            return sfName < ofName

        if sortColumn == NAME_COL_INDEX:
            return txt < otherTxt

        if sortColumn in [CALLERS_COL_INDEX, CALLEES_COL_INDEX]:
            return int(txt) < int(otherTxt)

        # Fallback to string comparison
        return txt < otherTxt

    def match(self, locationAndName):
        """Checks if the item function location and name
           matches what received
        """
        parts = locationAndName.split(":")
        if len(parts) != 3:
            return False
        return self.getFunctionName() == parts[2] and \
               str(self.getLineNumber()) == parts[1] and \
               self.getFileName() == parts[0]


class ProfilerTreeWidget(QTreeWidget):

    """Need only to generate sigEscapePressed signal"""

    sigEscapePressed = pyqtSignal()

    def __init__(self, parent=None):
        QTreeWidget.__init__(self, parent)

    def keyPressEvent(self, event):
        """Handles the key press events"""
        if event.key() == Qt.Key_Escape:
            self.sigEscapePressed.emit()
            event.accept()
        else:
            QTreeWidget.keyPressEvent(self, event)


class ProfileTableViewer(QWidget):

    """Profiling results table viewer"""

    sigEscapePressed = pyqtSignal()

    def __init__(self, scriptName, params, reportTime,
                 dataFile, stats, parent=None):
        QWidget.__init__(self, parent)

        self.__table = ProfilerTreeWidget(self)
        self.__table.sigEscapePressed.connect(self.__onEsc)

        self.__script = scriptName
        self.__stats = stats
        project = GlobalData().project
        if project.isLoaded():
            self.__projectPrefix = os.path.dirname(project.fileName)
        else:
            self.__projectPrefix = os.path.dirname(scriptName)
        if not self.__projectPrefix.endswith(os.path.sep):
            self.__projectPrefix += os.path.sep

        self.__table.setAlternatingRowColors(True)
        self.__table.setRootIsDecorated(False)
        self.__table.setItemsExpandable(False)
        self.__table.setSortingEnabled(True)
        self.__table.setItemDelegate(NoOutlineHeightDelegate(4))
        self.__table.setUniformRowHeights(True)
        self.__table.setSelectionMode(QAbstractItemView.SingleSelection)
        self.__table.setSelectionBehavior(QAbstractItemView.SelectRows)

        headerLabels = ["", "Calls", "Total time", "Per call",
                        "Cum. time", "Per call", "File name:line",
                        "Function", "Callers", "Callees"]
        self.__table.setHeaderLabels(headerLabels)

        headerItem = self.__table.headerItem()
        headerItem.setToolTip(0, "Indication if it is an outside function")
        headerItem.setToolTip(1, "Actual number of calls/primitive calls "
                              "(not induced via recursion)")
        headerItem.setToolTip(2, "Total time spent in function "
                              "(excluding time made in calls "
                              "to sub-functions)")
        headerItem.setToolTip(3, "Total time divided by number "
                              "of actual calls")
        headerItem.setToolTip(4, "Total time spent in function and all "
                              "subfunctions (from invocation till exit)")
        headerItem.setToolTip(5, "Cumulative time divided by number "
                              "of primitive calls")
        headerItem.setToolTip(6, "Function location")
        headerItem.setToolTip(7, "Function name")
        headerItem.setToolTip(8, "Function callers")
        headerItem.setToolTip(9, "Function callees")

        self.__table.itemActivated.connect(self.__activated)

        totalCalls = self.__stats.total_calls
        # The calls were not induced via recursion
        totalPrimitiveCalls = self.__stats.prim_calls
        totalTime = self.__stats.total_tt

        txt = "<b>Script:</b> " + self.__script + " " + \
              params['arguments'] + "<br/>" \
              "<b>Run at:</b> " + reportTime + "<br/>" + \
              str(totalCalls) + " function calls (" + \
              str(totalPrimitiveCalls) + " primitive calls) in " + \
              FLOAT_FORMAT % totalTime + " CPU seconds"
        summary = HeaderFitLabel(self)
        summary.setText(txt)
        summary.setToolTip(txt)
        summary.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Fixed)
        summary.setMinimumWidth(10)

        vLayout = QVBoxLayout()
        vLayout.setContentsMargins(0, 0, 0, 0)
        vLayout.setSpacing(0)
        vLayout.addWidget(summary)
        vLayout.addWidget(self.__table)

        self.setLayout(vLayout)
        self.__createContextMenu()

        self.__populate(totalTime)

    def __onEsc(self):
        """Triggered when Esc is pressed"""
        self.sigEscapePressed.emit()

    def __createContextMenu(self):
        """Creates context menu for the table raws"""
        self.__contextMenu = QMenu(self)
        self.__callersMenu = QMenu("Callers", self)
        self.__outsideCallersMenu = QMenu("Outside callers", self)
        self.__calleesMenu = QMenu("Callees", self)
        self.__outsideCalleesMenu = QMenu("Outside callees", self)
        self.__contextMenu.addMenu(self.__callersMenu)
        self.__contextMenu.addMenu(self.__outsideCallersMenu)
        self.__contextMenu.addSeparator()
        self.__contextMenu.addMenu(self.__calleesMenu)
        self.__contextMenu.addMenu(self.__outsideCalleesMenu)

        self.__callersMenu.triggered.connect(self.__onCallContextMenu)
        self.__outsideCallersMenu.triggered.connect(self.__onCallContextMenu)
        self.__calleesMenu.triggered.connect(self.__onCallContextMenu)
        self.__outsideCalleesMenu.triggered.connect(self.__onCallContextMenu)

        self.__table.setContextMenuPolicy(Qt.CustomContextMenu)
        self.__table.customContextMenuRequested.connect(self.__showContextMenu)

    def __showContextMenu(self, point):
        """Context menu"""
        self.__callersMenu.clear()
        self.__outsideCallersMenu.clear()
        self.__calleesMenu.clear()
        self.__outsideCalleesMenu.clear()

        # Detect what the item was clicked
        item = self.__table.itemAt(point)
        funcName = item.getFunctionName()

        # Build the context menu
        if item.callersCount() == 0:
            self.__callersMenu.setEnabled(False)
            self.__outsideCallersMenu.setEnabled(False)
        else:
            callers = self.__stats.stats[item.getFuncIDs()][4]
            callersList = list(callers.keys())
            callersList.sort()
            for callerFunc in callersList:
                menuText = self.__getCallLine(callerFunc, callers[callerFunc])
                if self.__isOutsideItem(callerFunc[0]):
                    act = self.__outsideCallersMenu.addAction(menuText)
                else:
                    act = self.__callersMenu.addAction(menuText)
                funcFileName, funcLine, funcName = \
                    self.__getLocationAndName(callerFunc)
                act.setData(funcFileName + ":" + str(funcLine) +
                            ":" + funcName)
            self.__callersMenu.setEnabled(not self.__callersMenu.isEmpty())
            self.__outsideCallersMenu.setEnabled(
                not self.__outsideCallersMenu.isEmpty())

        if item.calleesCount() == 0:
            self.__calleesMenu.setEnabled(False)
            self.__outsideCalleesMenu.setEnabled(False)
        else:
            callees = self.__stats.all_callees[item.getFuncIDs()]
            calleesList = list(callees.keys())
            calleesList.sort()
            for calleeFunc in calleesList:
                menuText = self.__getCallLine(calleeFunc, callees[calleeFunc])
                if self.__isOutsideItem(calleeFunc[0]):
                    act = self.__outsideCalleesMenu.addAction(menuText)
                else:
                    act = self.__calleesMenu.addAction(menuText)
                funcFileName, funcLine, funcName = \
                    self.__getLocationAndName(calleeFunc)
                act.setData(funcFileName + ":" +
                            str(funcLine) + ":" + funcName)
            self.__calleesMenu.setEnabled(not self.__calleesMenu.isEmpty())
            self.__outsideCalleesMenu.setEnabled(
                not self.__outsideCalleesMenu.isEmpty())

        self.__contextMenu.popup(QCursor.pos())

    def __resize(self):
        """Resizes columns to the content"""
        self.__table.header().resizeSections(QHeaderView.ResizeToContents)
        self.__table.header().setStretchLastSection(True)
        self.__table.header().resizeSection(OUTSIDE_COL_INDEX, 28)
        self.__table.header().setSectionResizeMode(OUTSIDE_COL_INDEX,
                                                   QHeaderView.Fixed)

    def setFocus(self):
        """Set focus to the proper widget"""
        self.__table.setFocus()

    def __isOutsideItem(self, fileName):
        """Detects if the record should be shown as an outside one"""
        return not fileName.startswith(self.__projectPrefix)

    def __activated(self, item, column):
        """Triggered when the item is activated"""
        try:
            line = item.getLineNumber()
            fileName = item.getFileName()
            if line == 0 or not os.path.isabs(fileName):
                return
            GlobalData().mainWindow.openFile(fileName, line)
        except:
            logging.error("Could not jump to function location")

    @staticmethod
    def __getFuncShortLocation(funcFileName, funcLine):
        """Provides unified shortened function location"""
        if funcFileName == "":
            return ""
        return os.path.basename(funcFileName) + ":" + str(funcLine)

    def __getCallLine(self, func, props):
        """Provides the formatted call line"""
        funcFileName, funcLine, funcName = self.__getLocationAndName(func)
        if isinstance(props, tuple):
            actualCalls, primitiveCalls, totalTime, cumulativeTime = props
            if primitiveCalls != actualCalls:
                callsString = str(actualCalls) + "/" + str(primitiveCalls)
            else:
                callsString = str(actualCalls)

            return callsString + "\t" + FLOAT_FORMAT % totalTime + "\t" + \
                   FLOAT_FORMAT % cumulativeTime + "\t" + \
                   self.__getFuncShortLocation(funcFileName, funcLine) + \
                   "(" + funcName + ")"

        # I've never seen this branch working so it is just in case.
        # There was something like this in the pstats module
        return self.__getFuncShortLocation(funcFileName, funcLine) + \
               "(" + funcName + ")"

    @staticmethod
    def __getLocationAndName(func):
        """Provides standardized function file name, line and its name"""
        if func[:2] == ('~', 0):
            # special case for built-in functions
            name = func[2]
            if name.startswith('<') and name.endswith('>'):
                return "", 0, "{%s}" % name[1:-1]
            return "", 0, name
        return func[0], func[1], func[2]

    def __createItem(self, func, totalCPUTime,
                     primitiveCalls, actualCalls, totalTime,
                     cumulativeTime, timePerCall, cumulativeTimePerCall,
                     callers):
        """Creates an item to display"""
        values = []
        values.append("")
        if primitiveCalls == actualCalls:
            values.append(str(actualCalls))
        else:
            values.append(str(actualCalls) + "/" + str(primitiveCalls))

        if totalCPUTime == 0.0:
            values.append(FLOAT_FORMAT % totalTime)
        else:
            values.append(FLOAT_FORMAT % totalTime +
                          " \t(%3.2f%%)" % (totalTime / totalCPUTime * 100))
        values.append(FLOAT_FORMAT % timePerCall)
        values.append(FLOAT_FORMAT % cumulativeTime)
        values.append(FLOAT_FORMAT % cumulativeTimePerCall)

        # Location and name will be filled in the item constructor
        values.append("")
        values.append("")

        # Callers
        callersCount = len(callers)
        values.append(str(callersCount))

        # Callees
        if func in self.__stats.all_callees:
            calleesCount = len(self.__stats.all_callees[func])
        else:
            calleesCount = 0
        values.append(str(calleesCount))

        item = ProfilingTableItem(values, self.__isOutsideItem(func[0]), func)

        if callersCount != 0:
            tooltip = ""
            callersList = list(callers.keys())
            callersList.sort()
            for callerFunc in callersList[:MAX_CALLS_IN_TOOLTIP]:
                if tooltip != "":
                    tooltip += "\n"
                tooltip += self.__getCallLine(callerFunc, callers[callerFunc])
            if callersCount > MAX_CALLS_IN_TOOLTIP:
                tooltip += "\n. . ."
            item.setToolTip(8, tooltip)

        if calleesCount != 0:
            callees = self.__stats.all_callees[func]
            tooltip = ""
            calleesList = list(callees.keys())
            calleesList.sort()
            for calleeFunc in calleesList[:MAX_CALLS_IN_TOOLTIP]:
                if tooltip != "":
                    tooltip += "\n"
                tooltip += self.__getCallLine(calleeFunc, callees[calleeFunc])
            if calleesCount > MAX_CALLS_IN_TOOLTIP:
                tooltip += "\n. . ."
            item.setToolTip(9, tooltip)

        self.__table.addTopLevelItem(item)

    def __populate(self, totalCPUTime):
        """Populates the data"""
        for func, (primitiveCalls, actualCalls, totalTime,
                   cumulativeTime, callers) in self.__stats.stats.items():
            # Calc time per call
            if actualCalls == 0:
                timePerCall = 0.0
            else:
                timePerCall = totalTime / actualCalls

            # Calc time per cummulative call
            if primitiveCalls == 0:
                cumulativeTimePerCall = 0.0
            else:
                cumulativeTimePerCall = cumulativeTime / primitiveCalls

            self.__createItem(func, totalCPUTime,
                              primitiveCalls, actualCalls, totalTime,
                              cumulativeTime, timePerCall,
                              cumulativeTimePerCall, callers)
        self.__resize()
        self.__table.header().setSortIndicator(2, Qt.DescendingOrder)
        self.__table.sortItems(2, self.__table.header().sortIndicatorOrder())

    def togglePath(self, state):
        """Switches between showing full paths or file names in locations"""
        for index in range(0, self.__table.topLevelItemCount()):
            self.__table.topLevelItem(index).updateLocation(state)
        self.__resize()
        return

    def __onCallContextMenu(self, act):
        """Triggered when a context menu action is requested"""
        name = str(act.data().toString())
        for index in range(0, self.__table.topLevelItemCount()):
            item = self.__table.topLevelItem(index)
            if item.match(name):
                self.__table.clearSelection()
                self.__table.scrollToItem(item)
                self.__table.setCurrentItem(item)

    def onSaveAs(self, fileName):
        """Saves the table to a file in CSV format"""
        try:
            f = open(fileName, "wt")
            headerItem = self.__table.headerItem()
            outsideIndex = -1
            for index in range(0, headerItem.columnCount()):
                title = str(headerItem.text(index))
                if title == "":
                    outsideIndex = index
                    title = "Outside"
                if index != 0:
                    f.write(',' + title)
                else:
                    f.write(title)

            for index in range(0, self.__table.topLevelItemCount()):
                item = self.__table.topLevelItem(index)
                f.write("\n")
                for column in range(0, item.columnCount()):
                    if column != 0:
                        f.write(',')
                    if column == outsideIndex:
                        if item.isOutside():
                            f.write("Y")
                        else:
                            f.write("N")
                    else:
                        text = str(item.text(column))
                        f.write(text.replace('\t', ''))
            f.close()
        except Exception as ex:
            logging.error(ex)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Client exceptions viewer"""


import os.path
from ui.qt import (Qt, pyqtSignal, QSize, QSizePolicy, QFrame, QTreeWidget,
                   QTreeWidgetItem, QVBoxLayout, QLabel, QWidget,
                   QAbstractItemView, QMenu, QHBoxLayout,
                   QCursor, QAction, QToolBar)
from ui.itemdelegates import NoOutlineHeightDelegate
from ui.labels import HeaderFitLabel
from ui.spacers import ToolBarExpandingSpacer
from utils.pixmapcache import getIcon
from utils.globals import GlobalData
from utils.project import CodimensionProject
from .variableitems import getDisplayValue, getTooltipValue


STACK_FRAME_ITEM = 0
EXCEPTION_ITEM = 1


class StackFrameItem(QTreeWidgetItem):

    """One stack trace frame"""

    def __init__(self, parentItem, fileName, lineNumber, funcName, funcArgs):
        QTreeWidgetItem.__init__(self, parentItem)

        self.__fileName = fileName
        self.__lineNumber = lineNumber
        self.__funcName = funcName
        self.__funcArgs = funcArgs
        self.setText(0, os.path.basename(fileName) + ":" + str(lineNumber))
        self.setToolTip(0, fileName + ":" + str(lineNumber))
        self.setText(1, funcName)
        self.setToolTip(1, funcName)
        self.setText(2, funcArgs)
        self.setToolTip(2, funcArgs)

    @staticmethod
    def getType():
        """Provides the item type"""
        return STACK_FRAME_ITEM

    def getLocation(self):
        """Provides the location in the code"""
        return self.toolTip(0)

    def getFileName(self):
        """Provides the file name"""
        return self.__fileName

    def getLineNumber(self):
        """Provides the line number"""
        return self.__lineNumber


class ExceptionItem(QTreeWidgetItem):

    """One exception item"""

    def __init__(self, parentItem, exceptionType, exceptionMessage,
                 stackTrace):
        QTreeWidgetItem.__init__(self, parentItem)
        self.__count = 1
        self.__exceptionType = exceptionType
        self.__exceptionMessage = exceptionMessage

        if not exceptionMessage:
            self.setText(0, str(exceptionType))
            self.setToolTip(0, "Type: " + str(exceptionType))
        else:
            self.setText(0, str(exceptionType) + ", " +
                         getDisplayValue(exceptionMessage))
            tooltip = "Type: " + str(exceptionType) + "\n" + "Message: "
            tooltipMessage = getTooltipValue(exceptionMessage)
            if '\r' in tooltipMessage or '\n' in tooltipMessage:
                tooltip += "\n" + tooltipMessage
            else:
                tooltip += tooltipMessage
            self.setToolTip(0, tooltip)

        if stackTrace:
            for entry in stackTrace:
                fileName = entry[0]
                lineNumber = entry[1]
                funcName = entry[2]
                funcArguments = entry[3]
                StackFrameItem(self, fileName, lineNumber,
                               funcName, funcArguments)

    @staticmethod
    def getType():
        """Provides the item type"""
        return EXCEPTION_ITEM

    def getCount(self):
        """Provides the number of same exceptions"""
        return self.__count

    def getExceptionType(self):
        """Provides the exception type"""
        return self.__exceptionType

    def incrementCounter(self):
        """Increments the counter of the same exceptions"""
        self.__count += 1
        if self.__exceptionMessage == "" or self.__exceptionMessage is None:
            self.setText(0, str(self.__exceptionType) +
                         " (" + str(self.__count) + " times)")
        else:
            self.setText(0, str(self.__exceptionType) +
                         " (" + str(self.__count) + " times), " +
                         getDisplayValue(self.__exceptionMessage))

    def equal(self, exceptionType, exceptionMessage, stackTrace):
        """Returns True if the exceptions are equal"""
        if exceptionType != self.__exceptionType:
            return False
        if exceptionMessage != self.__exceptionMessage:
            return False

        if stackTrace is None:
            stackTrace = []

        count = self.childCount()
        if count != len(stackTrace):
            return False

        for index in range(count):
            child = self.child(index)
            otherLocation = stackTrace[index][0] + ":" + \
                            str(stackTrace[index][1])
            if otherLocation != child.getLocation():
                return False
        return True


class ClientExceptionsViewer(QWidget):

    """Implements the client exceptions viewer for a debugger"""

    sigClientExceptionsCleared = pyqtSignal()

    def __init__(self, parent, ignoredExceptionsViewer):
        QWidget.__init__(self, parent)

        self.__ignoredExceptionsViewer = ignoredExceptionsViewer
        self.__currentItem = None

        self.__createPopupMenu()
        self.__createLayout()

        GlobalData().project.sigProjectChanged.connect(self.__onProjectChanged)

    def setFocus(self):
        """Sets the widget focus"""
        self.exceptionsList.setFocus()

    def __createPopupMenu(self):
        """Creates the popup menu"""
        self.__excptMenu = QMenu()
        self.__addToIgnoreMenuItem = self.__excptMenu.addAction(
            "Add to ignore list", self.__onAddToIgnore)
        self.__jumpToCodeMenuItem = self.__excptMenu.addAction(
            "Jump to code", self.__onJumpToCode)

    def __createLayout(self):
        """Creates the widget layout"""
        verticalLayout = QVBoxLayout(self)
        verticalLayout.setContentsMargins(0, 0, 0, 0)
        verticalLayout.setSpacing(0)

        self.__excptLabel = HeaderFitLabel(self)
        self.__excptLabel.setText('Exceptions')
        self.__excptLabel.setSizePolicy(QSizePolicy.Expanding,
                                        QSizePolicy.Fixed)
        self.__excptLabel.setMinimumWidth(10)

        self.headerToolbar = QToolBar(self)
        self.headerToolbar.setIconSize(QSize(18, 18))
        self.headerToolbar.setContentsMargins(1, 1, 1, 1)
        self.headerToolbar.addWidget(self.__excptLabel)

        self.exceptionsList = QTreeWidget(self)
        self.exceptionsList.setSortingEnabled(False)
        self.exceptionsList.setAlternatingRowColors(True)
        self.exceptionsList.setRootIsDecorated(True)
        self.exceptionsList.setItemsExpandable(True)
        self.exceptionsList.setUniformRowHeights(True)
        self.exceptionsList.setSelectionMode(QAbstractItemView.SingleSelection)
        self.exceptionsList.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.exceptionsList.setItemDelegate(NoOutlineHeightDelegate(4))
        self.exceptionsList.setContextMenuPolicy(Qt.CustomContextMenu)

        self.__addToIgnoreButton = QAction(
            getIcon('add.png'), "Add exception to the list of ignored", self)
        self.__addToIgnoreButton.triggered.connect(self.__onAddToIgnore)
        self.__addToIgnoreButton.setEnabled(False)

        self.__jumpToCodeButton = QAction(
            getIcon('gotoline.png'), "Jump to the code", self)
        self.__jumpToCodeButton.triggered.connect(self.__onJumpToCode)
        self.__jumpToCodeButton.setEnabled(False)

        self.__delAllButton = QAction(
            getIcon('trash.png'), "Delete all the client exceptions", self)
        self.__delAllButton.triggered.connect(self.__onDelAll)
        self.__delAllButton.setEnabled(False)

        self.toolbar = QToolBar()
        self.toolbar.setOrientation(Qt.Horizontal)
        self.toolbar.setMovable(False)
        self.toolbar.setAllowedAreas(Qt.TopToolBarArea)
        self.toolbar.setIconSize(QSize(16, 16))
        self.toolbar.setFixedHeight(28)
        self.toolbar.setContentsMargins(0, 0, 0, 0)
        self.toolbar.addAction(self.__addToIgnoreButton)
        self.toolbar.addAction(self.__jumpToCodeButton)
        self.toolbar.addWidget(ToolBarExpandingSpacer(self.toolbar))
        self.toolbar.addAction(self.__delAllButton)

        self.exceptionsList.itemDoubleClicked.connect(
            self.__onExceptionDoubleClicked)
        self.exceptionsList.customContextMenuRequested.connect(
            self.__showContextMenu)
        self.exceptionsList.itemSelectionChanged.connect(
            self.__onSelectionChanged)

        self.exceptionsList.setHeaderLabels(["Exception",
                                             "Function", "Arguments"])

        verticalLayout.addWidget(self.headerToolbar)
        verticalLayout.addWidget(self.toolbar)
        verticalLayout.addWidget(self.exceptionsList)

    def clear(self):
        """Clears the content"""
        self.exceptionsList.clear()
        self.__updateExceptionsLabel()
        self.__addToIgnoreButton.setEnabled(False)
        self.__jumpToCodeButton.setEnabled(False)
        self.__delAllButton.setEnabled(False)
        self.__currentItem = None
        self.sigClientExceptionsCleared.emit()

    def __onExceptionDoubleClicked(self, item, column):
        """Triggered when an exception is double clicked"""
        del item    # unused argument
        del column  # unused argument
        if self.__currentItem is not None:
            if self.__currentItem.getType() == STACK_FRAME_ITEM:
                self.__onJumpToCode()
                return

            # This is an exception item itself.
            # Open a separate dialog window with th detailed info.

    def __showContextMenu(self, coord):
        """Shows the frames list context menu"""
        self.__currentItem = self.exceptionsList.itemAt(coord)

        self.__addToIgnoreMenuItem.setEnabled(
            self.__addToIgnoreButton.isEnabled())
        self.__jumpToCodeMenuItem.setEnabled(
            self.__jumpToCodeButton.isEnabled())

        if self.__currentItem is not None:
            self.__excptMenu.popup(QCursor.pos())

    def __onAddToIgnore(self):
        """Adds an exception into the ignore list"""
        if self.__currentItem is not None:
            self.__ignoredExceptionsViewer.addExceptionFilter(
                str(self.__currentItem.getExceptionType()))
            self.__addToIgnoreButton.setEnabled(False)

    def __onJumpToCode(self):
        """Jumps to the corresponding source code line"""
        if self.__currentItem is not None:
            if self.__currentItem.getType() == STACK_FRAME_ITEM:
                fileName = self.__currentItem.getFileName()
                if '<' not in fileName and '>' not in fileName:
                    lineNumber = self.__currentItem.getLineNumber()

                    editorsManager = GlobalData().mainWindow.editorsManager()
                    editorsManager.openFile(fileName, lineNumber)
                    editor = editorsManager.currentWidget().getEditor()
                    editor.gotoLine(lineNumber)
                    editorsManager.currentWidget().setFocus()

    def __onDelAll(self):
        """Triggered when all the exceptions should be deleted"""
        self.clear()

    def addException(self, exceptionType, exceptionMessage, stackTrace):
        """Adds the exception to the view"""
        for index in range(self.exceptionsList.topLevelItemCount()):
            item = self.exceptionsList.topLevelItem(index)
            if item.equal(exceptionType, exceptionMessage, stackTrace):
                item.incrementCounter()
                self.exceptionsList.clearSelection()
                self.exceptionsList.setCurrentItem(item)
                self.__updateExceptionsLabel()
                return

        item = ExceptionItem(self.exceptionsList, exceptionType,
                             exceptionMessage, stackTrace)
        self.exceptionsList.clearSelection()
        self.exceptionsList.setCurrentItem(item)
        self.__updateExceptionsLabel()
        self.__delAllButton.setEnabled(True)

    def __updateExceptionsLabel(self):
        """Updates the exceptions header label"""
        total = self.getTotalCount()
        if total > 0:
            self.__excptLabel.setText("Exceptions (total: " + str(total) + ")")
        else:
            self.__excptLabel.setText("Exceptions")

    def getTotalCount(self):
        """Provides the total number of exceptions"""
        count = 0
        for index in range(self.exceptionsList.topLevelItemCount()):
            count += self.exceptionsList.topLevelItem(index).getCount()
        return count

    def __onProjectChanged(self, what):
        """Triggered when a project is changed"""
        if what == CodimensionProject.CompleteProject:
            self.clear()

    def __onSelectionChanged(self):
        """Triggered when the current item is changed"""
        selected = list(self.exceptionsList.selectedItems())
        if selected:
            self.__currentItem = selected[0]
            if self.__currentItem.getType() == STACK_FRAME_ITEM:
                fileName = self.__currentItem.getFileName()
                if '<' in fileName or '>' in fileName:
                    self.__jumpToCodeButton.setEnabled(False)
                else:
                    self.__jumpToCodeButton.setEnabled(True)
                self.__addToIgnoreButton.setEnabled(False)
            else:
                self.__jumpToCodeButton.setEnabled(False)
                excType = str(self.__currentItem.getExceptionType())
                if self.__ignoredExceptionsViewer.isIgnored(excType) or \
                   " " in excType or excType.startswith("unhandled"):
                    self.__addToIgnoreButton.setEnabled(False)
                else:
                    self.__addToIgnoreButton.setEnabled(True)
        else:
            self.__currentItem = None
            self.__addToIgnoreButton.setEnabled(False)
            self.__jumpToCodeButton.setEnabled(False)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4 and adopted for codimension.
# Original copyright:
# Copyright (c) 2007 - 2013 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""Module implementing the Watch expression model"""

from ui.qt import QAbstractItemModel, Qt, QModelIndex


class WatchPointModel(QAbstractItemModel):

    """Class implementing a custom model for watch expressions"""

    def __init__(self, parent=None):
        QAbstractItemModel.__init__(self, parent)

        self.watchpoints = []
        self.header = ['Condition', 'Special', 'Temporary', 'Enabled',
                       'Ignore Count']
        self.alignments = [Qt.Alignment(Qt.AlignLeft),
                           Qt.Alignment(Qt.AlignLeft),
                           Qt.Alignment(Qt.AlignHCenter),
                           Qt.Alignment(Qt.AlignHCenter),
                           Qt.Alignment(Qt.AlignRight)]

    def columnCount(self, parent=QModelIndex()):
        """Provides the current column count"""
        return len(self.header) + 1

    def rowCount(self, parent=QModelIndex()):
        """Provides the current row count"""
        # we do not have a tree, parent should always be invalid
        if not parent.isValid():
            return len(self.watchpoints)
        return 0

    def data(self, index, role):
        """Provides the requested data"""
        if not index.isValid():
            return None

        if role == Qt.DisplayRole or role == Qt.ToolTipRole:
            if index.column() < len(self.header):
                return self.watchpoints[index.row()][index.column()]

        if role == Qt.TextAlignmentRole:
            if index.column() < len(self.alignments):
                return self.alignments[index.column()]

        return None

    def flags(self, index):
        """Provides the item flags"""
        if not index.isValid():
            return Qt.ItemIsEnabled
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """Public method to get header data"""
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            if section >= len(self.header):
                return ""
            return self.header[section]
        return None

    def index(self, row, column, parent=QModelIndex()):
        """Creates the index"""
        if parent.isValid() or \
           row < 0 or row >= len(self.watchpoints) or \
           column < 0 or column >= len(self.header):
            return QModelIndex()
        return self.createIndex(row, column, self.watchpoints[row])

    def parent(self, index):
        """Provides the parent index"""
        return QModelIndex()

    def hasChildren(self, parent=QModelIndex()):
        """True if has children"""
        if not parent.isValid():
            return len(self.watchpoints) > 0
        return False

    def addWatchPoint(self, cond, special, properties):
        """Adds a new watch expression to the list"""
        wpoint = [cond, special] + list(properties)
        cnt = len(self.watchpoints)
        self.beginInsertRows(QModelIndex(), cnt, cnt)
        self.watchpoints.append(wpoint)
        self.endInsertRows()

    def setWatchPointByIndex(self, index, cond, special, properties):
        """Sets the values of a watch expression given by index"""
        if index.isValid():
            row = index.row()
            index1 = self.createIndex(row, 0, self.watchpoints[row])
            index2 = self.createIndex(row, len(self.watchpoints[row]),
                                      self.watchpoints[row])
            self.dataAboutToBeChanged.emit(index1, index2)
            i = 0
            for value in [cond, special] + list(properties):
                self.watchpoints[row][i] = value
                i += 1
            self.dataChanged.emit(index1, index2)

    def setWatchPointEnabledByIndex(self, index, enabled):
        """Sets the enabled state of a watch expression given by index"""
        if index.isValid():
            row = index.row()
            col = 3
            index1 = self.createIndex(row, col, self.watchpoints[row])
            self.dataAboutToBeChanged.emit(index1, index1)
            self.watchpoints[row][col] = enabled
            self.dataChanged.emit(index1, index1)

    def deleteWatchPointByIndex(self, index):
        """Sets the values of a watch expression given by index"""
        if index.isValid():
            row = index.row()
            self.beginRemoveRows(QModelIndex(), row, row)
            del self.watchpoints[row]
            self.endRemoveRows()

    def deleteWatchPoints(self, idxList):
        """Deletes a list of watch expressions given by their indexes"""
        rows = []
        for index in idxList:
            if index.isValid():
                rows.append(index.row())
        rows.sort(reverse=True)
        for row in rows:
            self.beginRemoveRows(QModelIndex(), row, row)
            del self.watchpoints[row]
            self.endRemoveRows()

    def deleteAll(self):
        """Deletes all watch expressions"""
        if self.watchpoints:
            self.beginRemoveRows(QModelIndex(), 0,
                                 len(self.watchpoints) - 1)
            self.watchpoints = []
            self.endRemoveRows()

    def getWatchPointByIndex(self, index):
        """Provides the values of a watch expression given by index"""
        if index.isValid():
            return self.watchpoints[index.row()][:] # return a copy
        return []

    def getWatchPointIndex(self, cond, special=""):
        """Provides the index of a watch expression given by expression"""
        for row in range(len(self.watchpoints)):
            wpoint = self.watchpoints[row]
            if wpoint[0] == cond:
                if special and wpoint[1] != special:
                    continue
                return self.createIndex(row, 0, self.watchpoints[row])
        return QModelIndex()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Version control system plugin interface"""

from .cdmpluginbase import CDMPluginBase


class VersionControlSystemInterface(CDMPluginBase):

    """Version control system plugin interface"""

    # Indicator for a non VCS item. It must be used in the getStatus(...)
    # report if the item is not under VCS control.
    NOT_UNDER_VCS = -1

    def __init__(self):
        """Plugin initialization.

        The plugin class is instantiated with no arguments.
        Instantiating is done regardless wheather a plugin is
        enabled or disabled. So it is recommended to have the
        resource allocation in the activate(...) method and
        deallocation of them in the deactivate(...) method.
        """
        CDMPluginBase.__init__(self)

    # Member functions below could or should be implemented by a plugin.
    # See docstrings for the detailed description.

    @staticmethod
    def isIDEVersionCompatible(ideVersion):
        """Codimension makes this call before activating a plugin.
           The passed ideVersion is a string representing
           the current IDE version.
           True should be returned if the plugin is compatible with the IDE."""
        raise Exception("isIDEVersionCompatible() must be overridden")

    @staticmethod
    def getVCSName():
        """Should provide the specific version control name, e.g. SVN"""
        raise Exception("getVCSName() must be overridden")

    def activate(self, ideSettings, ideGlobalData):
        """Override this method to do specific plugin activation handling.

        ideSettings - reference to the IDE Settings singleton
                      see codimension/src/utils/settings.py
        ideGlobalData - reference to the IDE global settings
                        see codimension/src/utils/globals.py

        Note: if overridden then call the base class activate() first.
              Plugin specific activation handling should follow it.
        """
        CDMPluginBase.activate(self, ideSettings, ideGlobalData)

    def deactivate(self):
        """Override this method to do specific plugin deactivation handling.

        Note: if overridden then first do the plugin specific deactivation
              handling and then call the base class deactivate()
        """
        CDMPluginBase.deactivate(self)

    def getConfigFunction(self):
        """A plugin may provide a function to be called for configuring it.

        If a plugin does not require any config parameters then None
        should be returned.
        By default no configuring is required.
        """
        return CDMPluginBase.getConfigFunction(self)

    def populateMainMenu(self, parentMenu):
        """A plugin may provide a main menu.

        The main menu looks as follows:
        Plugins
            - Plugin manager (fixed item)
            - Separator (visible if at lease one plugin provided its
              main menu)
            - <Plugin #1 name> (this is the parentMenu passed)
            ...
        If no items were populated by the plugin then there will be no
        <Plugin #N name> menu item shown.
        Codimension will remove the populated menu when
        a plugin is disabled.
        """
        del parentMenu  # unused argument

    def populateFileContextMenu(self, parentMenu):
        """A plugin may provide a file system file context menu.

        The file context menu shown in the project viewer window will have
        an item with a plugin name and subitems which are populated here.
        If no items were populated then the plugin menu item will not be shown.
        When a callback is called the parentMenu data will be set to the
        absolute path of the selected item. So the plugin will be able
        to retrieve the selected path as follows:
        # In this method
        self.fileParentMenu = parentMenu
        ...
        # In a menu item handler
        path = str( self.fileParentMenu.menuAction().data().toString() )

        Codimension will remove the populated menu when
        a plugin is disabled.
        """
        del parentMenu  # unused argument

    def populateDirectoryContextMenu(self, parentMenu):
        """A plugin may provide a file system directory context menu.

        The directory context menu shown in the project viewer window will
        have an item with a plugin name and subitems which are populated here.
        If no items were populated then the plugin menu item will not be shown.
        When a callback is called the parentMenu data will be set to the
        absolute path of the selected item. So the plugin will be able
        to retrieve the selected path as follows:
        # In this method
        self.dirParentMenu = parentMenu
        ...
        # In a menu item handler
        path = str( self.dirParentMenu.menuAction().data().toString() )

        Codimension will remove the populated menu when
        a plugin is disabled.
        """
        del parentMenu  # unused argument

    def populateBufferContextMenu(self, parentMenu):
        """A plugin may provide a current editor buffer context menu.

        The buffer context menu shown for the current edited/viewed file
        will have an item with a plugin name and subitems which are
        populated here.
        If no items were populated then the plugin menu item will not be
        shown.
        Codimension will remove the populated menu when a plugin is
        disabled.
        Note: when a buffer context menu is selected by the user it always
              refers to the current widget. To get access to the current
              editing widget the plugin can use:
              self.ide.currentEditorWidget
              The widget could be of different types and some circumstances
              should be considered, e.g.:
              - it could be a new file which has not been saved yet
              - it could be modified
              - it could be that the disk file has already been deleted
              - etc.
              Having the current widget reference the plugin is able to
              retrieve the information it needs.
        """
        del parentMenu  # unused argument

    def getCustomIndicators(self):
        """A plugin must provide a list of its custom indicators.

        Each indicator is a tuple:
        (id, pixmap, foreground, background, defaultTooltip)
        id - integer value which must be >= 0. Negative values are reserved
             for common indicators
        pixmap - QPixmap. It should be 16x16, if larger then the pixmap
                 will be scaled. The pixmap is used at the status bar.
        foreground - QColor or None. If not None then the color will be
                     used in the project browser for the text.
        background - QColor or None. If not None then the color will be used
                     to fill the project browser item background.
        defaultTooltip - string or None. If provided and the getStatus(...)
                         call did not provide a message then this one is
                         displayed
        """
        raise Exception("getCustomIndicators() must be overridden")

    # flag argement values for the getStatus() method
    REQUEST_ITEM_ONLY = 0
    REQUEST_DIRECTORY = 1
    REQUEST_RECURSIVE = 2

    def getStatus(self, path, flag):
        """A plugin should provide VCS statuses for the items.

        path - absolute path (string)
        flag - tells what is requested. It is one of the following:
          - REQUEST_ITEM_ONLY => the status of the path is requested
          - REQUEST_DIRECTORY => the path is a directory and the status of
                                 each item in the directory is requested
          - REQUEST_RECURSIVE => the path is a directory and the status of
                                 each item recursively is requested

        Return value:
          list of tuples for each reported item
          (rest-of-item-path, status, message) where
              - rest-of-item-path => the received path + this value
                                     must make the full item path
              - status => integer, whether NOT_UNDER_VCS or
                          a custom indicator identifier
              - message => message if so (string) or None. If a message
                           is supplied then it will appear as a tooltip
                           over the item indicator

        ATTENTION: the reported path MUST end with os.path.sep if it
                   was a directory.

        ATTENTION: no IDE functions must be called from this method. This
                   method is running in a separate thread and updating UI
                   from non-UI thread can cause the IDE crash.
        """
        del path    # unused argument
        del flag    # unused argument
        raise Exception("getStatus() must be overridden")



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Classes browser model"""

import os.path
from os.path import basename
from utils.fileutils import getFileProperties, isPythonMime
from utils.settings import Settings
from utils.project import CodimensionProject
from .viewitems import TreeViewClassItem
from .browsermodelbase import BrowserModelBase


class ClassesBrowserModel(BrowserModelBase):

    """Class implementing the project browser model"""

    def __init__(self, parent=None):
        BrowserModelBase.__init__(self, ["Name", "File name", "Line"], parent)
        self.setTooltips(Settings()['classesTooltips'])
        self.globalData.project.sigProjectChanged.connect(
            self.__onProjectChanged)

    def __populateModel(self):
        """Populates the project browser model"""
        self.clear()
        project = self.globalData.project
        cache = self.globalData.briefModinfoCache
        for fname in project.filesList:
            mime, _, _ = getFileProperties(fname)
            if isPythonMime(mime):
                info = cache.get(fname)
                for classObj in info.classes:
                    item = TreeViewClassItem(self.rootItem, classObj)
                    item.appendData([basename(fname), classObj.line])
                    item.setPath(fname)
                    self.rootItem.appendChild(item)

    def __onProjectChanged(self, what):
        """Triggered when a project is changed"""
        if what == CodimensionProject.CompleteProject:
            self.__populateModel()

    def onFSChanged(self, addedPythonFiles, deletedPythonFiles):
        """Triggered when some files appeared or disappeared"""
        needUpdate = False
        itemsToDelete = []
        for path in deletedPythonFiles:
            for item in self.rootItem.childItems:
                if os.path.realpath(path) == \
                   os.path.realpath(item.getPath()):
                    itemsToDelete.append(item)

        for item in itemsToDelete:
            needUpdate = True
            self.removeTreeItem(item)

        for path in addedPythonFiles:
            try:
                info = self.globalData.briefModinfoCache.get(path)
            except:
                # It could be that a file was created and deleted straight
                # away. In this case the cache will generate an exception.
                continue
            for classObj in info.classes:
                needUpdate = True
                newItem = TreeViewClassItem(self.rootItem, classObj)
                newItem.appendData([basename(path), classObj.line])
                newItem.setPath(path)
                self.addTreeItem(self.rootItem, newItem)
        return needUpdate

    def onFileUpdated(self, fileName):
        """Triggered when a file was updated"""
        # Here: python file which belongs to the project
        info = self.globalData.briefModinfoCache.get(fileName)

        existingClasses = []
        itemsToRemove = []
        needUpdate = False

        # For all root items
        path = os.path.realpath(fileName)
        for treeItem in self.rootItem.childItems:
            if os.path.realpath(treeItem.getPath()) != path:
                continue

            # Item belongs to the modified file
            name = treeItem.sourceObj.name
            found = False
            for cls in info.classes:
                if cls.name == name:
                    found = True
                    existingClasses.append(name)
                    treeItem.updateData(cls)
                    treeItem.setData(2, cls.line)
                    self.signalItemUpdated(treeItem)
                    self.updateSingleClassItem(treeItem, cls)
                    break
            if not found:
                itemsToRemove.append(treeItem)

        for item in itemsToRemove:
            needUpdate = True
            self.removeTreeItem(item)

        # Add those which have been introduced
        for item in info.classes:
            if item.name not in existingClasses:
                needUpdate = True
                newItem = TreeViewClassItem(self.rootItem, item)
                newItem.appendData([basename(fileName), item.line])
                newItem.setPath(fileName)
                self.addTreeItem(self.rootItem, newItem)
        return needUpdate



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy sergey.satskiy@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Redirected IO console messages implementation"""

from datetime import datetime
from utils.settings import Settings


def printableTimestamp(tstamp):
    """Provides a printable timestamp"""
    millisecond = str(int(round(tstamp.microsecond / 1000.0)))
    while len(millisecond) != 3:
        millisecond = "0" + millisecond
    return tstamp.strftime("%H:%M:%S.") + millisecond


def getNowTimestamp():
    """Provides the now() timestamp as a string"""
    return printableTimestamp(datetime.now())


class IOConsoleMsg():

    """Holds a single message"""

    __slots__ = ["msgType", "msgText", "timestamp"]

    IDE_MESSAGE = 0
    STDOUT_MESSAGE = 1
    STDERR_MESSAGE = 2
    STDIN_MESSAGE = 3

    def __init__(self, msgType, msgText):
        self.msgType = msgType
        self.msgText = msgText
        self.timestamp = datetime.now()

    def getTimestamp(self):
        """Provides the timestamp as a string"""
        return printableTimestamp(self.timestamp)


class IOConsoleMessages:

    """Holds a list of messages"""

    def __init__(self):
        self.msgs = []
        self.size = 0

    def append(self, msg):
        """Appends the message to the list. Returns True if it was trimmed"""
        self.msgs.append(msg)
        self.size += 1

        if self.size <= Settings()['ioconsolemaxmsgs']:
            return False

        removeCount = Settings()['ioconsoledelchunk']
        self.msgs = self.msgs[removeCount:]
        self.size -= removeCount
        return True

    def clear(self):
        """Clears all the messages"""
        self.msgs = []
        self.size = 0

    def renderWithTimestamps(self):
        """Provides the 'copy with timestamps' content"""
        buf = []
        outputIndex = -1

        streamPlaceholder = "   "
        timestampPlaceholder = " " * 12
        emptyPrefix = streamPlaceholder + " " + timestampPlaceholder

        for msg in self.msgs:
            if msg.msgType == IOConsoleMsg.IDE_MESSAGE:
                txt = msg.msgText.strip()
                parts = txt.splitlines()
                if parts:
                    buf.append("ide " + msg.getTimestamp() + " " + parts[0])
                    for part in parts[1:]:
                        buf.append(emptyPrefix + " " + part)
                outputIndex = -1
                continue

            if msg.msgType not in [IOConsoleMsg.STDOUT_MESSAGE,
                                   IOConsoleMsg.STDERR_MESSAGE,
                                   IOConsoleMsg.STDIN_MESSAGE]:
                raise Exception("Unexpected message IO console type: " +
                                str(msg.msgType))

            if msg.msgType == IOConsoleMsg.STDOUT_MESSAGE:
                stream = "out"
            elif msg.msgType == IOConsoleMsg.STDERR_MESSAGE:
                stream = "err"
            else:
                stream = " in"

            parts = msg.msgText.splitlines()
            if parts:
                if outputIndex == -1:
                    buf.append(stream + " " + msg.getTimestamp() +
                               " " + parts[0])
                    outputIndex = len(buf) - 1
                else:
                    spaces = len(buf[outputIndex]) - len(emptyPrefix) - 1
                    buf[outputIndex] += parts[0]
                    buf.append(stream + " " + msg.getTimestamp() +
                               " " + " " * spaces + "^")

                parts = parts[1:]
                if parts:
                    for part in parts:
                        buf.append(stream + " " + msg.getTimestamp() +
                                   " " + part)
                    outputIndex = len(buf) - 1

            if msg.msgText.endswith('\n') or msg.msgText.endswith('\r'):
                outputIndex = -1

        return "\n".join(buf)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2015-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Various comment items on a virtual canvas"""

# pylint: disable=C0305
# pylint: disable=R0902

from sys import maxsize
from ui.qt import (Qt, QPen, QBrush, QPainterPath, QGraphicsPathItem,
                   QGraphicsItem, QStyleOptionGraphicsItem, QStyle)
from .auxitems import Connector
from .cellelement import CellElement
from .routines import distance, getCommentBoxPath
from .textmixin import TextMixin



class CommentCellBase(CellElement, TextMixin):

    """Base class for comment items"""

    def __init__(self, ref, canvas, x, y):
        CellElement.__init__(self, ref, canvas, x, y)
        TextMixin.__init__(self)
        self._leftEdge = None
        self.connector = None

    def _copyToClipboard(self, parts):
        """Copies the item to a clipboard"""
        commonLeadingSpaces = maxsize
        for part in parts:
            commonLeadingSpaces = min(commonLeadingSpaces, part.beginPos - 1)

        content = []
        currentLine = parts[0].beginLine
        for part in parts:
            while part.beginLine - currentLine > 1:
                content.append('#')
                currentLine += 1

            content.append(part.getLineContent()[commonLeadingSpaces:])
            currentLine = part.beginLine
        self._putMimeToClipboard('\n'.join(content))

    def getSelectTooltip(self):
        """Provides the tooltip"""
        return 'Comment at ' + CellElement.getLinesSuffix(self.getLineRange())



class IndependentCommentCell(CommentCellBase, QGraphicsPathItem):

    """Represents a single independent comment"""

    def __init__(self, ref, canvas, x, y):
        CommentCellBase.__init__(self, ref, canvas, x, y)
        QGraphicsPathItem.__init__(self)
        self.kind = CellElement.INDEPENDENT_COMMENT
        self.leadingForElse = False
        self.sideForElse = False

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def render(self):
        """Renders the cell"""
        settings = self.canvas.settings
        self.setupText(self)

        self.minHeight = self.textRect.height() + \
                         2 * (settings.vCellPadding + settings.vTextPadding)
        self.minWidth = self.textRect.width() + \
                        2 * (settings.hCellPadding + settings.hTextPadding)
        self.minWidth = max(self.minWidth, settings.minWidth)
        self.height = self.minHeight
        self.width = self.minWidth
        return (self.width, self.height)

    def adjustWidth(self):
        """Used during rendering to adjust the width of the cell.

        The comment now can take some space on the left and the left hand
        side cell has to be rendered already.
        The width of this cell will take whatever is needed considering
        the comment shift to the left.
        """
        settings = self.canvas.settings
        cellToTheLeft = self.canvas.cells[self.addr[1]][self.addr[0] - 1]
        spareWidth = \
            cellToTheLeft.width - settings.mainLine - settings.hCellPadding
        boxWidth = self.textRect.width() + \
                   2 * (settings.hCellPadding + settings.hTextPadding)
        boxWidth = max(boxWidth, settings.minWidth)
        if spareWidth >= boxWidth:
            self.minWidth = 0
        else:
            self.minWidth = boxWidth - spareWidth
        self.width = self.minWidth

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY
        self.__setupPath()
        scene.addItem(self.connector)
        scene.addItem(self)

    def __setupPath(self):
        """Sets the path for painting"""
        settings = self.canvas.settings

        cellToTheLeft = self.canvas.cells[self.addr[1]][self.addr[0] - 1]
        self._leftEdge = \
            cellToTheLeft.baseX + settings.mainLine + settings.hCellPadding
        boxWidth = self.textRect.width() + \
                   2 * (settings.hCellPadding + settings.hTextPadding)
        boxWidth = max(boxWidth, settings.minWidth)
        path = getCommentBoxPath(settings, self._leftEdge, self.baseY,
                                 boxWidth, self.minHeight)
        self.setPath(path)

        # May be later the connector will look different for two cases below
        if self.leadingForElse:
            self.connector = Connector(
                self.canvas, self._leftEdge + settings.hCellPadding,
                self.baseY + self.minHeight / 2,
                cellToTheLeft.baseX + settings.mainLine,
                self.baseY + self.minHeight / 2)
        else:
            self.connector = Connector(
                self.canvas, self._leftEdge + settings.hCellPadding,
                self.baseY + self.minHeight / 2,
                cellToTheLeft.baseX + settings.mainLine,
                self.baseY + self.minHeight / 2)
        self.connector.penColor = settings.commentBorderColor
        self.connector.penWidth = settings.boxLineWidth

    def paint(self, painter, option, widget):
        """Draws the independent comment"""
        settings = self.canvas.settings
        self.setPen(self.getPainterPen(self.isSelected(),
                                       settings.commentBorderColor))
        self.setBrush(QBrush(settings.commentBGColor))

        # Hide the dotted outline
        itemOption = QStyleOptionGraphicsItem(option)
        if itemOption.state & QStyle.State_Selected != 0:
            itemOption.state = itemOption.state & ~QStyle.State_Selected
        QGraphicsPathItem.paint(self, painter, itemOption, widget)

        # Draw the text in the rectangle
        pen = QPen(settings.commentFGColor)
        painter.setFont(settings.monoFont)
        painter.setPen(pen)
        painter.drawText(
            self._leftEdge + settings.hCellPadding + settings.hTextPadding,
            self.baseY + settings.vCellPadding + settings.vTextPadding,
            self.textRect.width(), self.textRect.height(),
            Qt.AlignLeft, self.text)

    def mouseDoubleClickEvent(self, event):
        """Jump to the appropriate line in the text editor"""
        # Needed custom because this item is used for ifs 'else' side comment
        CellElement.mouseDoubleClickEvent(self, event,
                                          pos=self.ref.beginPos)

    def getLineRange(self):
        """Provides the line range"""
        return self.ref.getLineRange()

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        return self.ref.getAbsPosRange()

    def copyToClipboard(self):
        """Copies the item to a clipboard"""
        self._copyToClipboard(self.ref.parts)



class LeadingCommentCell(CommentCellBase, QGraphicsPathItem):

    """Represents a single leading comment"""

    def __init__(self, ref, canvas, x, y):
        CommentCellBase.__init__(self, ref, canvas, x, y)
        QGraphicsPathItem.__init__(self)
        self.kind = CellElement.LEADING_COMMENT

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def render(self):
        """Renders the cell"""
        settings = self.canvas.settings
        self.setupText(self,
                       customText=self.ref.leadingComment.getDisplayValue(),
                       customReplacement='')

        self.minHeight = \
            self.textRect.height() + \
            2 * settings.vCellPadding + 2 * settings.vTextPadding
        self.minWidth = \
            self.textRect.width() + \
            2 * settings.hCellPadding + 2 * settings.hTextPadding
        self.minWidth = max(self.minWidth, settings.minWidth)
        self.height = self.minHeight
        self.width = self.minWidth
        return (self.width, self.height)

    def adjustWidth(self):
        """Used during rendering to adjust the width of the cell.

        The comment now can take some space on the left and the left hand
        side cell has to be rendered already.
        The width of this cell will take whatever is needed considering
        the comment shift to the left
        """
        cellToTheLeft = self.canvas.cells[self.addr[1]][self.addr[0] - 1]
        if cellToTheLeft.kind != CellElement.CONNECTOR:
            # Not implemented yet
            return

        # Here: there is a connector on the left so we can move the comment
        #       safely
        settings = self.canvas.settings
        spareWidth = cellToTheLeft.width - cellToTheLeft.minWidth
        boxWidth = self.textRect.width() + \
                   2 * (settings.hCellPadding + settings.hTextPadding)
        boxWidth = max(boxWidth, settings.minWidth)
        if spareWidth >= boxWidth:
            self.minWidth = 0
        else:
            self.minWidth = boxWidth - spareWidth
        self.width = self.minWidth

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY
        self.__setupPath()
        scene.addItem(self.connector)
        scene.addItem(self)

    def __setupPath(self):
        """Sets the comment path"""
        settings = self.canvas.settings

        # Bottom adjustment
        yShift = self.height - self.minHeight
        baseY = self.baseY + yShift

        cellToTheLeft = self.canvas.cells[self.addr[1]][self.addr[0] - 1]
        if cellToTheLeft.kind != CellElement.CONNECTOR:
            # not implemented yet
            self._leftEdge = self.baseX
        else:
            self._leftEdge = \
                cellToTheLeft.baseX + \
                settings.mainLine + settings.hCellPadding
        boxWidth = self.textRect.width() + \
                   2 * (settings.hCellPadding + settings.hTextPadding)
        boxWidth = max(boxWidth, settings.minWidth)

        shift = self.hShift * 2 * settings.openGroupHSpacer
        self._leftEdge += shift
        path = getCommentBoxPath(settings, self._leftEdge, baseY,
                                 boxWidth, self.minHeight)
        self.setPath(path)

        self.connector = Connector(self.canvas, 0, 0, 0, 0)
        connectorPath = QPainterPath()
        connectorPath.moveTo(self._leftEdge + settings.hCellPadding,
                             baseY + self.minHeight / 2)
        connectorPath.lineTo(self._leftEdge,
                             baseY + self.minHeight / 2)
        connectorPath.lineTo(self._leftEdge - settings.hCellPadding,
                             baseY + self.minHeight + settings.vCellPadding)
        self.connector.setPath(connectorPath)
        self.connector.penColor = settings.commentBorderColor
        self.connector.penWidth = settings.boxLineWidth

        self._leftEdge -= shift

    def paint(self, painter, option, widget):
        """Draws the leading comment"""
        settings = self.canvas.settings
        self.setPen(self.getPainterPen(self.isSelected(),
                                       settings.commentBorderColor))
        self.setBrush(QBrush(settings.commentBGColor))

        # Bottom adjustment
        yShift = self.height - self.minHeight
        baseY = self.baseY + yShift

        # Left adjustments
        shift = self.hShift * 2 * settings.openGroupHSpacer
        self._leftEdge += shift

        # Hide the dotted outline
        itemOption = QStyleOptionGraphicsItem(option)
        if itemOption.state & QStyle.State_Selected != 0:
            itemOption.state = itemOption.state & ~QStyle.State_Selected
        QGraphicsPathItem.paint(self, painter, itemOption, widget)

        # Draw the text in the rectangle
        pen = QPen(settings.commentFGColor)
        painter.setFont(settings.monoFont)
        painter.setPen(pen)
        painter.drawText(
            self._leftEdge + settings.hCellPadding + settings.hTextPadding,
            baseY + settings.vCellPadding + settings.vTextPadding,
            self.textRect.width(), self.textRect.height(),
            Qt.AlignLeft, self.text)

        self._leftEdge -= shift

    def getLineRange(self):
        """Provides the line range"""
        return self.ref.leadingComment.getLineRange()

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        return self.ref.leadingComment.getAbsPosRange()

    def copyToClipboard(self):
        """Copies the item to a clipboard"""
        self._copyToClipboard(self.ref.leadingComment.parts)



class SideCommentCell(CommentCellBase, QGraphicsPathItem):

    """Represents a single side comment"""

    def __init__(self, ref, canvas, x, y):
        CommentCellBase.__init__(self, ref, canvas, x, y)
        QGraphicsPathItem.__init__(self)
        self.kind = CellElement.SIDE_COMMENT

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def getCommentText(self):
        """Provides the text"""
        cellToTheLeft = self.canvas.cells[self.addr[1]][self.addr[0] - 1]
        if cellToTheLeft.kind == CellElement.IMPORT:
            importRef = cellToTheLeft.ref
            linesBefore = self.ref.sideComment.beginLine - \
                          importRef.whatPart.beginLine
            if importRef.fromPart is not None:
                linesBefore += 1
        else:
            linesBefore = self.ref.sideComment.beginLine - \
                          self.ref.body.beginLine
        return '\n' * linesBefore + self.ref.sideComment.getDisplayValue()

    def render(self):
        """Renders the cell"""
        settings = self.canvas.settings
        self.setupText(self, customText=self.getCommentText(),
                       customReplacement='')

        self.minHeight = self.textRect.height() + \
            2 * settings.vCellPadding + 2 * settings.vTextPadding
        self.minWidth = self.textRect.width() + \
                        2 * settings.hCellPadding + 2 * settings.hTextPadding
        self.minWidth = max(self.minWidth, settings.minWidth)
        self.height = self.minHeight
        self.width = self.minWidth
        return (self.width, self.height)

    def adjustWidth(self):
        """Used during rendering to adjust the width of the cell.

        The comment now can take some space on the left and the left hand
        side cell has to be rendered already.
        The width of this cell will take whatever is needed considering
        the comment shift to the left.
        """
        settings = self.canvas.settings
        cellToTheLeft = self.canvas.cells[self.addr[1]][self.addr[0] - 1]
        spareWidth = cellToTheLeft.width - cellToTheLeft.minWidth
        boxWidth = self.textRect.width() + \
                   2 * (settings.hCellPadding + settings.hTextPadding)
        boxWidth = max(boxWidth, settings.minWidth)
        if spareWidth >= boxWidth:
            self.minWidth = 0
        else:
            self.minWidth = boxWidth - spareWidth
        self.width = self.minWidth

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY
        self.__setupPath()
        scene.addItem(self.connector)
        scene.addItem(self)

    def __setupPath(self):
        """Sets the comment path"""
        settings = self.canvas.settings

        cellToTheLeft = self.canvas.cells[self.addr[1]][self.addr[0] - 1]
        boxWidth = self.textRect.width() + \
                   2 * (settings.hCellPadding + settings.hTextPadding)
        boxWidth = max(boxWidth, settings.minWidth)
        if cellToTheLeft.kind == CellElement.CONNECTOR:
            ifCell = None
            for cell in self.canvas.cells[self.addr[1]]:
                if cell.kind == CellElement.IF:
                    ifCell = cell
                    break
        elif cellToTheLeft.kind == CellElement.IF:
            ifCell = cellToTheLeft
        else:
            ifCell = None

        if cellToTheLeft.kind == CellElement.CONNECTOR:
            # 'if' or 'elif' side comment when there is a connector
            self._leftEdge = \
                cellToTheLeft.baseX + settings.mainLine + settings.hCellPadding
            boxBaseY = self.baseY
            path = getCommentBoxPath(settings, self._leftEdge, boxBaseY,
                                     boxWidth, self.minHeight)

            width = 0
            index = self.addr[0] - 1
            while self.canvas.cells[self.addr[1]][index].kind == \
                  CellElement.CONNECTOR:
                width += self.canvas.cells[self.addr[1]][index].width
                index -= 1

            yPos = self.baseY + settings.vCellPadding + \
                   settings.ifSideCommentVShift

            self.connector = Connector(
                self.canvas, self._leftEdge + settings.hCellPadding,
                yPos,
                ifCell.baseX + ifCell.minWidth - settings.hCellPadding,
                yPos)
        else:
            # Regular box or 'if' without an rhs connector
            if ifCell:
                self._leftEdge = ifCell.baseX + ifCell.minWidth
            else:
                self._leftEdge = cellToTheLeft.baseX + cellToTheLeft.minWidth
            path = getCommentBoxPath(settings, self._leftEdge, self.baseY,
                                     boxWidth, self.minHeight)


            if ifCell:
                yPos = self.baseY + settings.vCellPadding + \
                       settings.ifSideCommentVShift
                self.connector = Connector(
                    self.canvas, self._leftEdge + settings.hCellPadding,
                    yPos,
                    ifCell.baseX + ifCell.minWidth - settings.hCellPadding,
                    yPos)
            else:
                height = min(self.minHeight / 2, cellToTheLeft.minHeight / 2)
                self.connector = Connector(
                    self.canvas, self._leftEdge + settings.hCellPadding,
                    self.baseY + height,
                    cellToTheLeft.baseX +
                    cellToTheLeft.minWidth - settings.hCellPadding,
                    self.baseY + height)

        self.connector.penColor = settings.commentBorderColor
        self.connector.penWidth = settings.boxLineWidth

        self.setPath(path)

    def paint(self, painter, option, widget):
        """Draws the side comment"""
        settings = self.canvas.settings
        self.setPen(self.getPainterPen(self.isSelected(),
                                       settings.commentBorderColor))
        self.setBrush(QBrush(settings.commentBGColor))

        # Hide the dotted outline
        itemOption = QStyleOptionGraphicsItem(option)
        if itemOption.state & QStyle.State_Selected != 0:
            itemOption.state = itemOption.state & ~QStyle.State_Selected
        QGraphicsPathItem.paint(self, painter, itemOption, widget)

        # Draw the text in the rectangle
        pen = QPen(settings.commentFGColor)
        painter.setFont(settings.monoFont)
        painter.setPen(pen)
        boxBaseY = self.baseY
        painter.drawText(
            self._leftEdge + settings.hCellPadding + settings.hTextPadding,
            boxBaseY + settings.vCellPadding + settings.vTextPadding,
            self.textRect.width(), self.textRect.height(),
            Qt.AlignLeft, self.text)

    def mouseDoubleClickEvent(self, event):
        """Jump to the appropriate line in the text editor"""
        CellElement.mouseDoubleClickEvent(self, event,
                                          pos=self.ref.sideComment.beginPos)

    def getLineRange(self):
        """Provides the line range"""
        return self.ref.sideComment.getLineRange()

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        return self.ref.sideComment.getAbsPosRange()

    def getDistance(self, absPos):
        """Provides a distance between the absPos and the item"""
        retval = maxsize
        for part in self.ref.sideComment.parts:
            # +1 is for finishing \n character
            dist = distance(absPos, part.begin, part.end + 1)
            if dist == 0:
                return 0
            retval = min(retval, dist)
        return retval

    def getLineDistance(self, line):
        """Provides a distance between the line and the item"""
        retval = maxsize
        for part in self.ref.sideComment.parts:
            dist = distance(line, part.beginLine, part.endLine)
            if dist == 0:
                return 0
            retval = min(retval, dist)
        return retval

    def copyToClipboard(self):
        """Copies the item to a clipboard"""
        self._copyToClipboard(self.ref.sideComment.parts)



class AboveCommentCell(CommentCellBase, QGraphicsPathItem):

    """Represents a single leading comment which is above certain blocks.

    Blocks are: try/except or for/else or while/else
    i.e. those which are scopes located in a single row
    """

    def __init__(self, ref, canvas, x, y):
        CommentCellBase.__init__(self, ref, canvas, x, y)
        QGraphicsPathItem.__init__(self)
        self.kind = CellElement.ABOVE_COMMENT
        self.needConnector = False
        self.commentConnector = None

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def render(self):
        """Renders the cell"""
        settings = self.canvas.settings
        self.setupText(self,
                       customText=self.ref.leadingComment.getDisplayValue(),
                       customReplacement='')

        self.minHeight = self.textRect.height() + \
                         2 * (settings.vCellPadding + settings.vTextPadding)
        # Width of the comment box itself
        self.minWidth = self.textRect.width() + \
                        2 * (settings.hCellPadding + settings.hTextPadding)
        self.minWidth = max(self.minWidth, settings.minWidth)

        # Add the connector space
        self.minWidth += settings.mainLine + settings.hCellPadding

        self.height = self.minHeight
        self.width = self.minWidth
        return (self.width, self.height)

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY
        self.__setupPath()

        if self.needConnector:
            settings = self.canvas.settings
            self.connector = Connector(
                self.canvas, baseX + settings.mainLine, baseY,
                baseX + settings.mainLine, baseY + self.height)
            scene.addItem(self.connector)

        scene.addItem(self.commentConnector)
        scene.addItem(self)

    def __setupPath(self):
        """Sets the comment path"""
        settings = self.canvas.settings

        # Bottom adjustment
        yShift = self.height - self.minHeight
        baseY = self.baseY + yShift

        self._leftEdge = \
            self.baseX + settings.mainLine + settings.hCellPadding
        boxWidth = self.textRect.width() + \
                   2 * (settings.hCellPadding + settings.hTextPadding)
        boxWidth = max(boxWidth, settings.minWidth)

        path = getCommentBoxPath(settings, self._leftEdge, baseY,
                                 boxWidth, self.minHeight)
        self.setPath(path)

        self.commentConnector = Connector(self.canvas, 0, 0, 0, 0)
        connectorPath = QPainterPath()
        connectorPath.moveTo(self._leftEdge + settings.hCellPadding,
                             baseY + self.minHeight / 2)
        connectorPath.lineTo(self._leftEdge,
                             baseY + self.minHeight / 2)
        connectorPath.lineTo(self._leftEdge - settings.hCellPadding,
                             baseY + self.minHeight + settings.vCellPadding)
        self.commentConnector.setPath(connectorPath)
        self.commentConnector.penColor = settings.commentBorderColor
        self.commentConnector.penWidth = settings.boxLineWidth

    def paint(self, painter, option, widget):
        """Draws the leading comment"""
        settings = self.canvas.settings
        self.setPen(self.getPainterPen(self.isSelected(),
                                       settings.commentBorderColor))
        self.setBrush(QBrush(settings.commentBGColor))

        # Bottom adjustment
        yShift = self.height - self.minHeight
        baseY = self.baseY + yShift

        # Hide the dotted outline
        itemOption = QStyleOptionGraphicsItem(option)
        if itemOption.state & QStyle.State_Selected != 0:
            itemOption.state = itemOption.state & ~QStyle.State_Selected
        QGraphicsPathItem.paint(self, painter, itemOption, widget)

        # Draw the text in the rectangle
        pen = QPen(settings.commentFGColor)
        painter.setFont(settings.monoFont)
        painter.setPen(pen)
        painter.drawText(
            self._leftEdge + settings.hCellPadding + settings.hTextPadding,
            baseY + settings.vCellPadding + settings.vTextPadding,
            self.textRect.width(), self.textRect.height(),
            Qt.AlignLeft, self.text)

    def adjustWidth(self):
        """No adjustment needed"""

    def getLineRange(self):
        """Provides the line range"""
        return self.ref.leadingComment.getLineRange()

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        return self.ref.leadingComment.getAbsPosRange()

    def copyToClipboard(self):
        """Copies the item to a clipboard"""
        self._copyToClipboard(self.ref.leadingComment.parts)




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


#
# The file was taken from eric 4.4.3 and adopted for codimension.
# Original copyright:
# Copyright (c) 2007 - 2010 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""Splash screen implementation"""

from utils.pixmapcache import getPixmap
from .qt import (Qt, QColor, QApplication, QSplashScreen)


class SplashScreen(QSplashScreen):

    """Splash screen class"""

    def __init__(self):
        self.labelAlignment = Qt.Alignment(Qt.AlignBottom |
                                           Qt.AlignRight |
                                           Qt.AlignAbsolute)

        # The window flags are needed for some X Servers. E.g. Xwin-32
        # on windows draws a normal window outline if the flags are not here
        QSplashScreen.__init__(self, None, getPixmap('splash.png'),
                               Qt.SplashScreen |
                               Qt.WindowStaysOnTopHint |
                               Qt.X11BypassWindowManagerHint)

        self.show()
        QApplication.flush()

    def showMessage(self, msg):
        """Show the message in the bottom part of the splashscreen"""
        QSplashScreen.showMessage(self, msg,
                                  self.labelAlignment, QColor(Qt.black))
        QApplication.processEvents()

    def clearMessage(self):
        """Clear the splash screen message"""
        QSplashScreen.clearMessage(self)
        QApplication.processEvents()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy sergey.satskiy@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""VCS plugin support:
   manager to keep track of the VCS plugins and file status"""

import os.path
import logging
import datetime
from utils.settings import Settings
from utils.globals import GlobalData
from utils.project import CodimensionProject
from ui.qt import QObject, QTimer, pyqtSignal
from plugins.categories.vcsiface import VersionControlSystemInterface
from .statuscache import VCSStatusCache, VCSStatus
from .indicator import VCSIndicator
from .vcspluginthread import VCSPluginThread, IND_VCS_ERROR


class VCSPluginDescriptor(QObject):

    """Holds information about a single active plugin"""

    def __init__(self, manager, pluginID, plugin):
        QObject.__init__(self)

        self.manager = manager
        self.pluginID = pluginID
        self.plugin = plugin
        self.thread = None                  # VCSPluginThread
        self.indicators = {}                # ID -> VCSIndicator

        self.__getPluginIndicators()
        self.thread = VCSPluginThread(plugin)
        self.thread.start()

        self.thread.VCSStatus.connect(self.__onStatus)

    def stopThread(self):
        """Stops the plugin thread synchronously"""
        self.thread.stop()  # Sends request
        self.thread.wait()  # Joins the thread

    def requestStatus(self, path, flag, urgent=False):
        """Requests the item status asynchronously"""
        self.thread.addRequest(path, flag, urgent)

    def clearRequestQueue(self):
        """Clears the thread request queue"""
        self.thread.clearRequestQueue()

    def getPluginName(self):
        """Safe plugin name"""
        try:
            return self.plugin.getName()
        except:
            return "Unknown (could not retrieve)"

    def __getPluginIndicators(self):
        """Retrieves indicators from the plugin"""
        try:
            for indicatorDesc in self.plugin.getObject().getCustomIndicators():
                try:
                    indicator = VCSIndicator(indicatorDesc)
                    if indicator.identifier < 0:
                        logging.error("Custom VCS plugin '" +
                                      self.getPluginName() +
                                      "' indicator identifier " +
                                      str(indicator.identifier) +
                                      " is invalid. It must be >= 0. "
                                      "Ignore and continue.")
                    else:
                        self.indicators[indicator.identifier] = indicator
                except Exception as exc:
                    logging.error("Error getting custom VCS plugin '" +
                                  self.getPluginName() +
                                  "' indicator: " + str(exc))
        except Exception as exc:
            logging.error("Error getting custom indicators for a VCS plugin " +
                          self.getPluginName() + ". Exception: " +
                          str(exc))

    def __onStatus(self, path, indicatorId, message):
        """Triggered when the thread reported a status"""
        self.manager.updateStatus(path, self.pluginID, indicatorId, message)

    def canInitiateStatusRequestLoop(self):
        """Returns true if the is no jam in the request queue"""
        return self.thread.canInitiateStatusRequestLoop()


class VCSManager(QObject):

    """Manages the VCS plugins"""

    sigVCSFileStatus = pyqtSignal(str, object)
    sigVCSDirStatus = pyqtSignal(str, object)

    def __init__(self):
        QObject.__init__(self)

        self.dirCache = VCSStatusCache()    # Path -> VCSStatus
        self.fileCache = VCSStatusCache()   # Path -> VCSStatus
        self.activePlugins = {}             # Plugin ID -> VCSPluginDescriptor
        self.systemIndicators = {}          # ID -> VCSIndicator

        self.__firstFreeIndex = 0

        self.__readSettingsIndicators()
        self.__dirRequestLoopTimer = QTimer(self)
        self.__dirRequestLoopTimer.setSingleShot(True)
        self.__dirRequestLoopTimer.timeout.connect(
            self.__onDirRequestLoopTimer)

        GlobalData().project.sigProjectChanged.connect(self.__onProjectChanged)
        GlobalData().project.sigFSChanged.connect(self.__onFSChanged)
        GlobalData().pluginManager.sigPluginActivated.connect(
            self.__onPluginActivated)

        # Plugin deactivation must be done via dismissPlugin(...)
        return

    def __getNewPluginIndex(self):
        """Provides a new plugin index"""
        index = self.__firstFreeIndex
        self.__firstFreeIndex += 1
        return index

    def __readSettingsIndicators(self):
        """Reads the system indicators"""
        for indicLine in Settings()['vcsindicators']:
            indicator = VCSIndicator(indicLine)
            self.systemIndicators[indicator.identifier] = indicator

    def __onPluginActivated(self, plugin):
        """Triggered when a plugin is activated"""
        if plugin.categoryName != "VersionControlSystemInterface":
            return

        newPluginIndex = self.__getNewPluginIndex()
        self.activePlugins[newPluginIndex] = VCSPluginDescriptor(
            self, newPluginIndex, plugin)
        plugin.getObject().PathChanged.connect(self.__onPathChanged)

        # Need to send requests for the opened TAB files
        mainWindow = GlobalData().mainWindow
        editorsManager = mainWindow.editorsManagerWidget.editorsManager
        editorsManager.sendAllTabsVCSStatusRequest()

        if self.activePluginCount() == 1 and GlobalData().project.isLoaded():
            # This is the first plugin and a project is there
            self.__populateProjectDirectories()
        self.__sendDirectoryRequestsOnActivation(newPluginIndex)
        self.__sendFileRequestsOnActivation(newPluginIndex)

        if self.activePluginCount() == 1:
            self.__startDirRequestTimer()

    def __onPathChanged(self, path):
        """The way plugins signal that a path has been changed"""
        # What is essentially required is to update the status of the path.
        # setLocallyModified(...) will do - it sends urgent request
        self.setLocallyModified(path)

        # Changed paths may change the file contents
        mainWindow = GlobalData().mainWindow
        editorsManager = mainWindow.editorsManagerWidget.editorsManager
        editorsManager.checkOutsidePathChange(path)

    def __populateProjectDirectories(self):
        """Populates the project directories in the dirCache"""
        project = GlobalData().project
        for path in project.filesList:
            if path.endswith(os.path.sep):
                self.dirCache.updateStatus(path, None, None, None, None)

    def __onProjectChanged(self, what):
        """Triggered when a project has changed"""
        if what == CodimensionProject.CompleteProject:
            self.dirCache.clear()
            self.fileCache.clear()
            if len(self.activePlugins) == 0:
                return

            # There are some plugins
            for _, descriptor in self.activePlugins.items():
                descriptor.clearRequestQueue()

            if GlobalData().project.isLoaded():
                self.__populateProjectDirectories()

    def __onFSChanged(self, items):
        """Triggered when files/dirs in the project are changed"""
        for item in items:
            item = str(item)
            if item.startswith('-'):
                item = item[1:]
                if item.endswith(os.path.sep):
                    # Directory removed
                    if item in self.dirCache.cache:
                        del self.dirCache.cache[item]
                else:
                    # File deleted
                    if item in self.fileCache.cache:
                        del self.fileCache.cache[item]
                    for _, descriptor in self.activePlugins.items():
                        descriptor.requestStatus(
                            item,
                            VersionControlSystemInterface.REQUEST_ITEM_ONLY,
                            True)
            else:
                item = item[1:]
                if item.endswith(os.path.sep):
                    # Directory added
                    if item not in self.dirCache.cache:
                        self.dirCache.updateStatus(
                            item, None, None, None, None)
                    for _, descriptor in self.activePlugins.items():
                        descriptor.requestStatus(
                            item,
                            VersionControlSystemInterface.REQUEST_DIRECTORY)
                else:
                    # File added
                    for _, descriptor in self.activePlugins.items():
                        descriptor.requestStatus(
                            item,
                            VersionControlSystemInterface.REQUEST_ITEM_ONLY,
                            True)

    def __sendDirectoryRequestsOnActivation(self, pluginID):
        """Sends the directory requests to the given plugin"""
        descriptor = self.activePlugins[pluginID]
        statuses = [None, VersionControlSystemInterface.NOT_UNDER_VCS]
        for path, status in self.dirCache.cache.items():
            if status.indicatorID in statuses:
                descriptor.requestStatus(
                    path, VersionControlSystemInterface.REQUEST_DIRECTORY)

    def __sendFileRequestsOnActivation(self, pluginID):
        """Sends the file requests to the given plugin"""
        descriptor = self.activePlugins[pluginID]
        statuses = [None, VersionControlSystemInterface.NOT_UNDER_VCS]
        for path, status in self.fileCache.cache.items():
            if status.indicatorID in statuses:
                descriptor.requestStatus(
                    path, VersionControlSystemInterface.REQUEST_ITEM_ONLY)

    def __sendPeriodicDirectoryRequests(self):
        """Sends the directory requests periodically"""
        for path, status in self.dirCache.cache.items():
            if status.pluginID in self.activePlugins:
                # This directory is under certain VCS, send the request
                # only to this VCS
                descriptor = self.activePlugins[status.pluginID]
                descriptor.requestStatus(
                    path, VersionControlSystemInterface.REQUEST_DIRECTORY)
            else:
                # The directory is not under any known VCS. Send the request
                # to all the registered VCS plugins
                for _, descriptor in self.activePlugins.items():
                    descriptor.requestStatus(
                        path, VersionControlSystemInterface.REQUEST_DIRECTORY)

    def dismissAllPlugins(self):
        """Stops all the plugin threads"""
        self.__dirRequestLoopTimer.stop()

        for identifier, descriptor in self.activePlugins.items():
            descriptor.plugin.getObject().PathChanged.disconnect(
                self.__onPathChanged)
            descriptor.stopThread()

        self.dirCache.clear()
        self.fileCache.clear()
        self.activePlugins = {}

    def dismissPlugin(self, plugin):
        """Stops the plugin thread and cleans the plugin data"""
        pluginID = None
        for identifier, descriptor in self.activePlugins.items():
            if descriptor.getPluginName() == plugin.getName():
                descriptor.plugin.getObject().PathChanged.disconnect(
                    self.__onPathChanged)
                pluginID = identifier
                descriptor.stopThread()
                self.fileCache.dismissPlugin(
                    pluginID, self.sendFileStatusNotification)
                self.dirCache.dismissPlugin(
                    pluginID, self.sendDirStatusNotification)

        if pluginID is not None:
            del self.activePlugins[pluginID]

        if self.activePluginCount() == 0:
            self.__dirRequestLoopTimer.stop()

    def requestStatus(self, path,
                      flag=VersionControlSystemInterface.REQUEST_ITEM_ONLY):
        """Provides the path status asynchronously via sending a signal"""
        delta = datetime.timedelta(0, Settings()['vcsstatusupdateinterval'])
        if path.endswith(os.path.sep):
            status = self.dirCache.getStatus(path)
            if status:
                self.sendDirStatusNotification(path, status)
            else:
                self.sendDirStatusNotification(path, VCSStatus())
        else:
            status = self.fileCache.getStatus(path)
            if status:
                self.sendFileStatusNotification(path, status)
            else:
                self.sendFileStatusNotification(path, VCSStatus())

        if status is None:
            for _, descriptor in self.activePlugins.items():
                descriptor.requestStatus(path, flag, True)
        else:
            if status.indicatorID is None or \
               status.lastUpdate is None or \
               datetime.datetime.now() - status.lastUpdate > delta:
                # Outdated or never been received
                if status.pluginID in self.activePlugins:
                    # Path is claimed by a plugin
                    descriptor = self.activePlugins[status.pluginID]
                    descriptor.requestStatus(path, flag, True)
                    return
                # Path is not claimed by any plugin - send to all
                for _, descriptor in self.activePlugins.items():
                    descriptor.requestStatus(path, flag, True)

    def setLocallyModified(self, path):
        """Sets the item status as locally modified"""
        for _, descriptor in self.activePlugins.items():
            descriptor.requestStatus(
                path, VersionControlSystemInterface.REQUEST_ITEM_ONLY, True)

    def sendDirStatusNotification(self, path, status):
        """Sends a signal that a status of the directory is changed"""
        self.sigVCSDirStatus.emit(path, status)

    def sendFileStatusNotification(self, path, status):
        """Sends a signal that a status of the file is changed"""
        self.sigVCSFileStatus.emit(path, status)

    def updateStatus(self, path, pluginID, indicatorID, message):
        """Called when a plugin thread reports a status"""
        if path.endswith(os.path.sep):
            self.dirCache.updateStatus(
                path, pluginID, indicatorID, message,
                self.sendDirStatusNotification)
        else:
            self.fileCache.updateStatus(
                path, pluginID, indicatorID, message,
                self.sendFileStatusNotification)
        # print("Status of " + path + " is " + str(indicatorID) +
        # " Message: " + str(message) + " Plugin ID: " + str(pluginID))

    def activePluginCount(self):
        """Returns the number of active VCS plugins"""
        return len(self.activePlugins)

    def drawStatus(self, vcsLabel, status):
        """Draw the VCS status"""
        if status.pluginID not in self.activePlugins:
            vcsLabel.setVisible(False)
            return

        descriptor = self.activePlugins[status.pluginID]
        if status.indicatorID not in descriptor.indicators:
            # Check the standard indicator
            if status.indicatorID in self.systemIndicators:
                indicator = self.systemIndicators[status.indicatorID]
                indicator.draw(vcsLabel)
                if status.message:
                    vcsLabel.setToolTip(status.message)
                else:
                    vcsLabel.setToolTip(indicator.defaultTooltip)
            else:
                # Neither plugin, no standard indicator
                try:
                    indicator = self.systemIndicators[IND_VCS_ERROR]
                    indicator.draw(vcsLabel)
                    vcsLabel.setToolTip("VCS plugin provided undefined "
                                        "indicator (id is " +
                                        str(status.indicatorID) + ")")
                except:
                    # No way to display indicator
                    vcsLabel.setVisible(False)
            return

        indicator = descriptor.indicators[status.indicatorID]
        indicator.draw(vcsLabel)
        if status.message:
            vcsLabel.setToolTip(status.message)
        else:
            vcsLabel.setToolTip(indicator.defaultTooltip)

    def getStatusIndicator(self, status):
        """Provides the VCS status indicator description for the given status.
           It is mostly used for the project browser"""
        if status.pluginID not in self.activePlugins:
            return None

        descriptor = self.activePlugins[status.pluginID]
        if status.indicatorID not in descriptor.indicators:
            # Check the standard indicator
            return self.getSystemIndicator(status.indicatorID)

        return descriptor.indicators[status.indicatorID]

    def getSystemIndicator(self, indicatorID):
        """Provides the IDE defined indicator if so"""
        if indicatorID in self.systemIndicators:
            return self.systemIndicators[indicatorID]
        return None

    def __canInitiateStatusRequestLoop(self):
        """Returns true if the is no jam in the request queue"""
        for _, descriptor in self.activePlugins.items():
            if descriptor.canInitiateStatusRequestLoop():
                return True
        return False

    def __onDirRequestLoopTimer(self):
        """Triggered when the dir request loop timer is fired"""
        self.__dirRequestLoopTimer.stop()
        if self.activePluginCount() == 0:
            return

        if self.__canInitiateStatusRequestLoop():
            self.__sendPeriodicDirectoryRequests()
        self.__startDirRequestTimer()

    def __startDirRequestTimer(self):
        """Starts the periodic request timer"""
        self.__dirRequestLoopTimer.start(
            Settings()['vcsstatusupdateinterval'] * 1000)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2016-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""QT color and font general utils"""

from ui.qt import QColor, QFont, QFontComboBox
from .globals import GlobalData
from .settings import Settings


def checkColorRange(value):
    """Checks the color range"""
    if value < 0 or value > 255:
        raise Exception("Invalid color value")


def toHex(value):
    """Converts the value to a double digit hex string"""
    asStr = hex(value)[2:]
    if len(asStr) == 1:
        return '0' + asStr
    return asStr


def buildColor(color):
    """Six options are supported:

    #hhh                hexadecimal rgb
    #hhhh               hexadecimal rgba
    #hhhhhh             hexadecimal rrggbb
    #hhhhhhhh           hexadecimal rrggbbaa
    ddd,ddd,ddd         decimal rgb
    ddd,ddd,ddd,ddd     decimal rgba
    """
    if color.startswith('#'):
        def normalizeLength(spec):
            if len(spec) in [6, 8]:
                return spec
            normalized = ''
            for character in spec:
                normalized += 2 * character
            return normalized

        color = color[1:]
        length = len(color)
        if length not in [3, 4, 6, 8]:
            raise Exception("Invalid hexadecimal color format: #" + color)

        try:
            # The most common case
            normColor = normalizeLength(color)
            red = int(normColor[0:2], 16)
            checkColorRange(red)
            green = int(normColor[2:4], 16)
            checkColorRange(green)
            blue = int(normColor[4:6], 16)
            checkColorRange(blue)

            if length in [3, 6]:
                return QColor(red, green, blue)
            alpha = int(normColor[6:8], 16)
            checkColorRange(alpha)
            return QColor(red, green, blue, alpha)
        except:
            raise Exception("Invalid hexadecimal color format: #" + color)

    parts = color.split(',')
    length = len(parts)
    if length not in [3, 4]:
        raise Exception("Invalid decimal color format: " + color)

    try:
        red = int(parts[0].strip())
        checkColorRange(red)
        green = int(parts[1].strip())
        checkColorRange(green)
        blue = int(parts[2].strip())
        checkColorRange(blue)

        if length == 3:
            return QColor(red, green, blue)
        alpha = int(parts[3].strip())
        checkColorRange(alpha)
        return QColor(red, green, blue, alpha)
    except:
        raise Exception("Invalid decimal color format: " + color)


def cssLikeColor(color):
    """Converts to a css like string, possibly shortened"""
    asStr = color.name()
    alpha = color.alpha()
    if alpha != 255:
        asStr += toHex(alpha)

    # Shorten it if possible
    if len(asStr) == 7:
        # '#rgb'
        if asStr[1] == asStr[2] and asStr[3] == asStr[4] and \
           asStr[5] == asStr[6]:
            return '#' + asStr[1] + asStr[3] + asStr[5]
        return asStr

    # '#rgba'
    if asStr[1] == asStr[2] and asStr[3] == asStr[4] and \
       asStr[5] == asStr[6] and asStr[7] == asStr[8]:
        return '#' + asStr[1] + asStr[3] + asStr[5] + asStr[7]
    return asStr


def transparentColor(fgColor, bgColor, alpha):
    """Converts the color value considering the background and alpha"""
    if alpha == 255:
        return fgColor

    dAlpha = float(alpha)/255.0
    dFirst = float(fgColor) * dAlpha
    dSecond = float(bgColor) * (1.0 - dAlpha)
    return min(255, round(dFirst + dSecond))


def qtCssColor(color, bgColor=None):
    """Converts the given color to the QT compatible one (transparency is not supported"""
    alpha = color.alpha()
    if alpha == 255:
        return color.name()

    if bgColor is None:
        rTransp = transparentColor(color.red(), 255, alpha)
        gTransp = transparentColor(color.green(), 255, alpha)
        bTransp = transparentColor(color.blue(), 255, alpha)
    else:
        rTransp = transparentColor(color.red(), bgColor.red(), alpha)
        gTransp = transparentColor(color.green(), bgColor.green(), alpha)
        bTransp = transparentColor(color.blue(), bgColor.blue(), alpha)
    return ''.join(('#', toHex(rTransp), toHex(gTransp), toHex(bTransp)))


def colorAsString(color, hexadecimal=False):
    """Converts the given color to a string"""
    if hexadecimal:
        return '#' + ''.join([toHex(color.red()),
                              toHex(color.green()),
                              toHex(color.blue()),
                              toHex(color.alpha())])
    return ','.join([str(color.red()),
                     str(color.green()),
                     str(color.blue()),
                     str(color.alpha())])


def buildFont(fontAsStr):
    """Converts saved font into QFont object"""
    fontAsStr = fontAsStr.strip()
    font = QFont()
    font.fromString(fontAsStr)
    return font


def fontAsString(font):
    """Converts a font to a string"""
    return font.toString()


excludeFontList = ['webdings', 'cursor', 'mathematical', 'dingbats', 'hershey']


def isExcludeFont(family):
    """True if the font needs to be excluded"""
    lowerFamily = family.lower()
    for exclusion in excludeFontList:
        if exclusion in lowerFamily:
            return True
    return False


def getFontList(fontType):
    """Provides a list of certain type fonts"""
    result = []
    combo = QFontComboBox()
    combo.setFontFilters(fontType)
    for index in range(combo.count()):
        family = str(combo.itemText(index))
        if not isExcludeFont(family):
            result.append(family)
    return result


def getMonospaceFontList():
    """Provides a list of strings with the system installed monospace fonts"""
    return getFontList(QFontComboBox.MonospacedFonts)


def getScalableFontList():
    """Provides a list of scalable fonts"""
    return getFontList(QFontComboBox.ScalableFonts)


def getProportionalFontList():
    """Provides a list of proportional fonts"""
    return getFontList(QFontComboBox.ProportionalFonts)


def colorFontToJSON(pythonObj):
    """Custom serialization"""
    if isinstance(pythonObj, QColor):
        return colorAsString(pythonObj)
    if isinstance(pythonObj, QFont):
        return fontAsString(pythonObj)
    raise TypeError(repr(pythonObj) + ' is not JSON serializable')


def colorFontFromJSON(jsonObj):
    """Custom deserialization"""
    if '__class__' in jsonObj:
        if jsonObj['__class__'] == 'QColor':
            return buildColor(jsonObj['__value__'])
        if jsonObj['__class__'] == 'QFont':
            return buildFont(jsonObj['__value__'])
    return jsonObj


def getZoomedMonoFont():
    """Provides the current mono font respecting zoom"""
    font = QFont(GlobalData().skin['monoFont'])
    font.setPointSize(font.pointSize() + Settings()['zoom'])
    return font


def getZoomedCFMonoFont():
    """Provides the current mono font respecting zoom"""
    font = QFont(GlobalData().skin['cfMonoFont'])
    font.setPointSize(font.pointSize() + Settings()['flowZoom'])
    return font


def getZoomedCFBadgeFont():
    """Provides the current control flow badge font respecting zoom"""
    font = QFont(GlobalData().skin['badgeFont'])
    font.setPointSize(font.pointSize() + Settings()['flowZoom'])
    return font


def getZoomedMarginFont():
    """Provides the current margin font respecting zoom"""
    font = QFont(GlobalData().skin['lineNumFont'])
    font.setPointSize(font.pointSize() + Settings()['zoom'])
    return font


def setLineEditBackground(widget, color, bgColor=None):
    """Sets the widget background to the given color"""
    widget.setStyleSheet('QLineEdit {background-color: ' +
                         qtCssColor(color, bgColor) + '}')

